# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any

from conf_ini_g.phase.specification.parameter.type import type_t
from conf_ini_g.phase.typed.parameter.invalid import INVALID_VALUE, invalid_value_t
from conf_ini_g.phase.typed.parameter.unit import ConvertedValue


def TypedValue(
    value: str,
    expected_type: type_t,
    /,
    *,
    unit: str | None = None,
    units: dict[str, int | float | invalid_value_t] = None,
) -> tuple[Any, list[str]]:
    """
    With unit consumed or not.
    """
    final_value, issues = expected_type.InterpretedValueOf(value)
    if issues.__len__() > 0:
        return INVALID_VALUE, [f"{value}: Invalid value: {', '.join(issues)}"]

    if (units is None) or (unit is None):
        return final_value, []

    conversion_factor = units.get(unit, None)
    if conversion_factor is None:
        return INVALID_VALUE, [f"{unit}: Missing unit definition."]
    if conversion_factor is INVALID_VALUE:
        return INVALID_VALUE, [f"{unit}: Invalid unit value."]

    converted, unconverted = ConvertedValue(final_value, conversion_factor)
    if unconverted.__len__() > 0:
        unconverted = ", ".join(unconverted)
        return INVALID_VALUE, [
            f"{unconverted}: Value(s) do(es) not support unit conversion."
        ]

    return converted, []



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Any

from babelwidget.main import backend_t
from babelwidget.main import dropdown_choice_h as dropdown_choice_wgt_h

from conf_ini_g.catalog.specification.annotation.choices import choices_t
from conf_ini_g.interface.window.parameter.value import value_wgt_a
from conf_ini_g.phase.specification.parameter.type import type_t

INDEX_FOR_NONE = 0


@dtcl.dataclass(repr=False, eq=False)
class choices_wgt_t(value_wgt_a):
    library_wgt: dropdown_choice_wgt_h

    @classmethod
    def NewForSpecification(
        cls,
        stripe: type_t | choices_t,
        backend: backend_t,
        /,
    ) -> choices_wgt_t:
        """
        If stripe does not contain the necessary details, the initial value (if valid) is the only choice, or a unique
        default choice ending with an exclamation point is added.
        """
        output = cls(library_wgt=backend.dropdown_choice_t())

        if isinstance(stripe, type_t):
            annotation = stripe.FirstAnnotationWithType(choices_t)
            if annotation is None:
                raise ValueError("No choices provided.")
        else:
            annotation = stripe

        for choice in annotation.options:
            output.library_wgt.addItem(choice)

        return output

    def Assign(self, value: str | None, _: Any, /) -> None:
        """"""
        if value is None:
            where = INDEX_FOR_NONE
        else:
            choices = tuple(map(self.itemText, range(self.count())))
            try:
                where = choices.index(value)
            except ValueError:
                choices = " or ".join(choices)
                raise ValueError(f"Invalid value: Actual={value}; Expected={choices}.")

        self.setCurrentIndex(where)

    def Text(self) -> str:
        """"""
        return self.library_wgt.currentText()

    def __getattr__(self, attribute: str, /) -> Any:
        """
        E.g., used for "SetFunction".
        """
        try:
            output = self.__getattribute__(attribute)
        except AttributeError:
            output = getattr(self.library_wgt, attribute)

        return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from enum import Enum as enum_t
from pathlib import Path as pl_path_t
from typing import Annotated

from str_to_obj import annotation_t
from str_to_obj.type.hint import annotated_hint_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class path_t(annotation_t):
    class TARGET_TYPE(enum_t):
        document = 1
        folder = 2
        any = 3

    ACCEPTED_TYPES = (pl_path_t,)
    target_type: enum_t
    # TODO: Change this into input, output, or both.
    is_input: bool

    @classmethod
    def NewAnnotatedType(
        cls, target_type: TARGET_TYPE, is_input: bool, /
    ) -> annotated_hint_t:
        """"""
        return Annotated[pl_path_t, cls(target_type=target_type, is_input=is_input)]

    def ValueIsCompliant(self, value: pl_path_t | None, /) -> list[str]:
        """
        None: Unspecified path.
        """
        issues = annotation_t.ValueIsCompliant(self, value)
        if issues.__len__() > 0:
            return issues

        if value is None:
            return []

        if not self.is_input:
            return []

        if value.exists():
            if self.target_type is path_t.TARGET_TYPE.any:
                if value.is_file() or value.is_dir():
                    return []
                else:
                    return [f"{value}: Not a valid file or folder."]

            if (self.target_type is path_t.TARGET_TYPE.document) and value.is_file():
                return []

            if (self.target_type is path_t.TARGET_TYPE.folder) and value.is_dir():
                return []

        return [
            f"{value}: Non-existent file or folder, or file for folder, or folder for file."
        ]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from configparser import ConfigParser as config_t
from configparser import ExtendedInterpolation

from conf_ini_g.extension.path import any_path_h
from conf_ini_g.phase.untyped.config import ini_config_h

_DEFAULT_OPTIONS = (
    ("delimiters", ("=",)),
    ("comment_prefixes", ("#",)),
    ("inline_comment_prefixes", ("#",)),
    ("interpolation", ExtendedInterpolation()),
)


def NewConfigFromPath(path: any_path_h, /, **kwargs) -> ini_config_h:
    """"""
    options = kwargs.copy()
    for name, value in _DEFAULT_OPTIONS:
        if name not in options:
            options[name] = value

    config = config_t(**options)
    config.optionxform = lambda option: option  # To avoid automatic lowercasing
    config.read(path)

    # config.sections(): Does not include the default section named configparser.DEFAULTSECT
    return {_nme: dict(config[_nme]) for _nme in config.sections()}
    # return {
    #     _nme: {
    #         _prm: _vle for _prm, _vle in config[_nme].items()
    #     }
    #     for _nme in config.sections()
    # }



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl

from issue_manager import ISSUE_MANAGER

from conf_ini_g.phase.specification.base import base_t
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.section.controller import controller_t


@dtcl.dataclass(repr=False, eq=False)
class section_t(base_t):
    category: str = "Main"
    optional: bool = False
    is_growable: bool = False  # True by default if no parameters are specified.
    parameters: list[parameter_t] | None = None
    controller: controller_t = None
    alternatives: dict[str, list[parameter_t]] = None

    def __post_init__(self) -> None:
        """"""
        if self.parameters is None:
            self.is_growable = True

        _SignalIssues(self)

    @property
    def controlling_values(self) -> tuple[str, ...]:
        """
        Call only on controlled sections.
        """
        if self.controller.primary_value is None:
            # The controlling values are not mentioned in the specification. They will
            # be set programmatically later on.
            return ()

        if self.alternatives is None:
            # The controlling values are not mentioned in the specification. They are
            # being set programmatically.
            return (self.controller.primary_value,)

        return (self.controller.primary_value,) + tuple(self.alternatives.keys())


def _SignalIssues(self) -> None:
    """"""
    with ISSUE_MANAGER.AddedContextLevel(
        f'Specification of {type(self).__name__} "{self.name}"'
    ):
        if self.parameters is not None:
            valid_name_sets = [tuple(_prm.name for _prm in self.parameters)]
            if self.alternatives is not None:
                for parameters in self.alternatives.values():
                    valid_name_sets.append(tuple(_prm.name for _prm in parameters))
            for valid_name_set in valid_name_sets:
                if valid_name_set.__len__() > set(valid_name_set).__len__():
                    ISSUE_MANAGER.Add(
                        "Repeated parameter names (possibly in alternatives)"
                    )

        basic = self.basic
        optional = self.optional

        if not (basic or optional):
            ISSUE_MANAGER.Add("Section is not basic but not optional")

        if (
            (self.controller is None)
            and (not optional)
            and ((self.parameters is None) or (self.parameters.__len__() == 0))
        ):
            ISSUE_MANAGER.Add("Empty mandatory section")

        if self.parameters is not None:
            n_parameters = 0
            n_basic_prms = 0
            for parameter in self.parameters:
                n_parameters += 1
                if parameter.basic:
                    n_basic_prms += 1

                if parameter.basic and not basic:
                    ISSUE_MANAGER.Add(
                        f'Basic parameter "{parameter.name}" in advanced section'
                    )
                if optional and not parameter.optional:
                    ISSUE_MANAGER.Add(
                        f'Mandatory parameter "{parameter.name}" in optional section'
                    )

            if (n_parameters == 0) and not self.is_growable:
                ISSUE_MANAGER.Add("Empty section not accepting runtime parameters")
            if basic and (n_parameters > 0) and (n_basic_prms == 0):
                ISSUE_MANAGER.Add("Basic section without any basic parameters")

        if self.controller is None:
            if self.alternatives is not None:
                ISSUE_MANAGER.Add("Uncontrolled section with alternatives")
        else:
            if self.controller.section == self.name:
                ISSUE_MANAGER.Add("Section controlled by itself")
            controlling_values = self.controlling_values
            if controlling_values.__len__() > set(controlling_values).__len__():
                ISSUE_MANAGER.Add(
                    "Controlled section with duplicated controlling values"
                )
            if self.parameters is not None:
                for parameter in self.parameters:
                    if not parameter.optional:
                        ISSUE_MANAGER.Add(
                            f'Mandatory parameter "{parameter.name}" in '
                            f'controlled section'
                        )

            if self.controller.primary_value is None:
                if self.alternatives is not None:
                    ISSUE_MANAGER.Add("Controlled section without primary value")
                # else: This will be set programmatically later on.
            elif self.alternatives is None:
                ISSUE_MANAGER.Add("Controlled section without alternatives")



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from conf_ini_g.interface.storage.parameter import INI_UNIT_SEPARATOR


def ValueUnitAndComment(
    value_w_unit_w_comment: str,
    comment_marker: str,
    /,
) -> tuple[str, str | None, str | None]:
    """"""
    value_w_unit, comment = _Pieces(value_w_unit_w_comment, comment_marker)
    if (comment is not None) and (comment.__len__() == 0):
        comment = None

    value_as_str, unit = _Pieces(value_w_unit, INI_UNIT_SEPARATOR, from_left=False)
    # if unit.__len__() == 0, do not make it None so that an invalid unit error is noticed later on

    return value_as_str, unit, comment


def _Pieces(
    combined: str, separator: str, /, *, from_left: bool = True
) -> tuple[str, str | None]:
    """"""
    if from_left:
        where_separator = combined.find(separator)
    else:
        where_separator = combined.rfind(separator)

    if where_separator != -1:
        left = combined[:where_separator].strip()
        right = combined[(where_separator + separator.__len__()) :].strip()
    else:
        left = combined
        right = None

    return left, right



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import configparser as cfpr
import sys as sstm

from str_to_obj import ObjectFromStr

from conf_ini_g.extension.path import ValidateOutputPath, any_path_h, path_t
from conf_ini_g.interface.storage.parameter import (
    INI_COMMENT_MARKER,
    INI_VALUE_ASSIGNMENT,
)
from conf_ini_g.phase.generic.config import AsStr, any_dict_config_h
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.section.generic import section_t
from conf_ini_g.phase.specification.section.unit import IsUnitSection
from conf_ini_g.phase.untyped.config import ini_config_h

_SECTIONS = "SECTIONS"


def DraftSpecificationFromINIDocument(path: any_path_h, /) -> str | None:
    """"""
    ini_config = INIConfig(path)
    if ini_config is None:
        return None

    sections = []
    for section_name, parameters in ini_config.items():
        # possibly_fuzzy=True: in case the raw config is not valid in that respect
        if IsUnitSection(section_name, possibly_fuzzy=True):
            continue

        parameters_as_lst = []
        for parameter_name, value_as_str in parameters.items():
            value, _ = ObjectFromStr(value_as_str)
            parameter = (
                f"{parameter_t.__name__}(\n"
                f'                name="{parameter_name}",\n'
                f"                type={type(value).__name__},\n"
                f"            )"
            )
            parameters_as_lst.append(parameter)

        parameters_as_str = ",\n            ".join(parameters_as_lst)
        section = (
            f"    {section_t.__name__}(\n"
            f'        name="{section_name}",\n'
            f"        parameters=[\n"
            f"            {parameters_as_str}\n"
            f"        ]\n"
            f"    )"
        )
        sections.append(section)

    imports = (
        f"# To use this specification file:\n"
        f"#     1. import the object {_SECTIONS}\n"
        f"#     2. instantiate a conf_ini_g.specification.config.config_t from it\n"
        f"from conf_ini_g.specification.parameter.main "
        f"import {parameter_t.__name__}\n"
        f"from conf_ini_g.specification.section.main import {section_t.__name__}\n"
    )

    return imports + f"\n{_SECTIONS} = (\n" + ",\n".join(sections) + ",\n)\n"


def INIConfig(ini_path: any_path_h | None, /) -> ini_config_h:
    """"""
    if ini_path is None:
        ini_config = {}
    else:
        ini_config = cfpr.ConfigParser(
            delimiters=INI_VALUE_ASSIGNMENT,
            comment_prefixes=INI_COMMENT_MARKER,
            empty_lines_in_values=False,
            interpolation=None,
        )
        ini_config.optionxform = lambda option: option
        # Returns DEFAULT <Section: DEFAULT> if path does not exist or is a folder.
        ini_config.read(ini_path, encoding=sstm.getfilesystemencoding())

    output = {
        section: {parameter: value for parameter, value in parameters.items()}
        for section, parameters in ini_config.items()
        if section != cfpr.DEFAULTSECT
    }

    return output


def SaveConfigToINIDocument(
    config: any_dict_config_h,
    path: any_path_h,
    /,
    *,
    should_overwrite: bool = False,
    should_raise_on_error: bool = False,
) -> list[str]:
    """"""
    path = path_t(path)
    issues = ValidateOutputPath(
        path,
        should_overwrite=should_overwrite,
        should_raise_on_error=should_raise_on_error,
    )
    if issues.__len__() > 0:
        return issues

    encoding = sstm.getfilesystemencoding()

    with path.open("w", encoding=encoding) as ini_writer:
        print(AsStr(config), file=ini_writer)
        print(f"", file=ini_writer)

    return []


# if __name__ == "__main__":
#     """
#     Run with: python -m conf_ini_g.interface.storage.config
#     from package base folder.
#     """
#     main_encoding = sstm.getfilesystemencoding()
#
#     parser = argument_parser_t(
#         prog="python -m conf_ini_g.interface.storage.config",
#         description="Display or save a draft config specification "
#         "based on an INI file.",
#         allow_abbrev=False,
#     )
#     # type=argparse.FileType() => automatic file opening
#     parser.add_argument(
#         dest="ini_path",
#         help="Input: Path to INI configuration file",
#         metavar="INI_config_file",
#     )
#     # type=argparse.FileType('w', encoding=main_encoding) => automatic creation of file
#     parser.add_argument(
#         dest="draft_specification",
#         help="Output: Path to draft config specification file. "
#         "If not passed, specification is displayed in console.",
#         default=None,
#         nargs="?",
#         metavar="draft_spec_file",
#     )
#     parser.add_argument(
#         "--overwrite",
#         dest="should_overwrite",
#         action="store_true",
#         help="Allows overwriting of draft config specification file.",
#     )
#
#     arguments = parser.parse_args()
#     ini_path = path_t(arguments.ini_path)
#     draft_path = arguments.draft_specification
#     main_should_overwrite = arguments.should_overwrite
#
#     issues_ = ValidateInputPath(ini_path)
#     if issues_.__len__() > 0:
#         print("\n".join(issues_) + "\n")
#         parser.print_help()
#         sstm.exit(-1)
#
#     if draft_path is not None:
#         draft_path = path_t(draft_path)
#         issues_ = ValidateOutputPath(draft_path, should_overwrite=main_should_overwrite)
#         if issues_.__len__() > 0:
#             print("\n".join(issues_) + "\n")
#             parser.print_help()
#             sstm.exit(-1)
#
#     draft = DraftSpecificationFromINIDocument(ini_path)
#     if draft is None:
#         print(f"{ini_path}: Not a valid INI document")
#         sstm.exit(-1)
#
#     if draft_path is None:
#         print(draft)
#     else:
#         with draft_path.open("w", encoding=main_encoding) as draft_writer:
#             draft_writer.write(draft)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from types import UnionType

from babelwidget.main import backend_t
from babelwidget.main import base_h as library_wgt_h
from babelwidget.main import dropdown_choice_h as dropdown_choice_wgt_h
from babelwidget.main import label_h as label_wgt_h
from babelwidget.main import text_line_h as text_line_wgt_h
from str_to_obj.type.hint_tree import hint_tree_t

from conf_ini_g.catalog.interface.window.parameter.directory import (
    ValueWidgetTypeForType,
)
from conf_ini_g.catalog.interface.window.parameter.multitype import multitype_wgt_t
from conf_ini_g.interface.storage.parameter import INI_UNIT_SEPARATOR
from conf_ini_g.interface.window.generic import FormattedName
from conf_ini_g.interface.window.parameter.type import TypeSelector
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(repr=False, eq=False)
class parameter_t:
    """
    In order to leave the section widget put the name, type, and input widgets of each parameter in columns,
    parameter_t is not a container widget. Instead, it just stores its component widgets for later addition to a layout.
    """

    spec_name: str
    name: label_wgt_h = dtcl.field(init=False, default=None)
    type: label_wgt_h | dropdown_choice_wgt_h = dtcl.field(init=False, default=None)
    value: library_wgt_h = dtcl.field(init=False, default=None)
    unit: text_line_wgt_h = dtcl.field(init=False, default=None)
    comment: str = dtcl.field(init=False, default=None)

    @classmethod
    def NewForSpecification(
        cls,
        parameter_spec: parameter_spec_t,
        backend: backend_t,
        /,
    ) -> parameter_t:
        """"""
        output = cls(spec_name=parameter_spec.name)

        formatted_name = FormattedName(parameter_spec.name, " ")
        comment = (
            f"{formatted_name}\n{parameter_spec.definition}.\n\n"
            f"{parameter_spec.description}."
        )
        output.name = backend.label_t(formatted_name)
        output.name.setToolTip(comment)
        output.comment = comment

        stripe = parameter_spec.type
        output.type, output.value = TypeAndValueWidgetsForType(stripe, backend)

        output.unit = backend.text_line_t()

        name_style = "padding-right: 5px;"
        if parameter_spec.optional:
            name_style += "color: gray;"
        output.name.setStyleSheet(name_style)
        output.type.setStyleSheet(name_style)

        return output

    def SetVisible(self, visible: bool, /) -> None:
        """"""
        self.name.setVisible(visible)
        self.type.setVisible(visible)
        self.value.library_wgt.setVisible(visible)
        if self.unit is not None:
            self.unit.setVisible(visible)

    def Text(self) -> str:
        """"""
        text = self.value.Text()
        if self.unit is None:
            return text

        unit = self.unit.Text()
        if unit.__len__() > 0:
            return f"{text}{INI_UNIT_SEPARATOR}{unit}"

        return text


def TypeAndValueWidgetsForType(
    stripe: type_t | hint_tree_t, backend: backend_t, /
) -> tuple[type[library_wgt_h], type[library_wgt_h]]:
    """"""
    if isinstance(stripe, hint_tree_t) and (stripe.type is UnionType):
        hint_options = stripe.elements
        type_wgt = TypeSelector(hint_options, backend)
        value_wgt = multitype_wgt_t.NewForHints(
            hint_options,
            type_wgt,
            backend,
        )
    else:
        type_wgt = backend.label_t(stripe.template_as_str)
        widget_type = ValueWidgetTypeForType(stripe)
        value_wgt = widget_type.NewForSpecification(
            stripe,
            backend,
        )

    return type_wgt, value_wgt



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import textwrap as text
from typing import Any, Sequence

from issue_manager import ISSUE_MANAGER
from rich.text import Text as text_t
from str_to_obj import ObjectFromStr
from str_to_obj.interface.console import TypeAsRichStr
from str_to_obj.type.hint import any_hint_h

from conf_ini_g.catalog.specification.annotation.callable import callable_t
from conf_ini_g.catalog.specification.annotation.choices import choices_t
from conf_ini_g.extension.python import SpecificationPath
from conf_ini_g.interface.storage.parameter import INI_UNIT_SEPARATOR
from conf_ini_g.interface.storage.section import INI_UNIT_SECTION
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.type import ANY_TYPE, type_t
from conf_ini_g.phase.specification.parameter.unit import unit_t
from conf_ini_g.phase.specification.parameter.value import MISSING_REQUIRED_VALUE
from conf_ini_g.phase.specification.section.controller import controller_t
from conf_ini_g.phase.specification.section.generic import section_t
from conf_ini_g.phase.specification.section.specific import (
    any_section_h,
    controlled_section_t,
    free_section_t,
    parameters_t,
)
from conf_ini_g.phase.specification.section.unit import IsUnitSection


@dtcl.dataclass(slots=True, repr=False, eq=False)
class config_t(list[any_section_h]):
    sections: dtcl.InitVar[Sequence[section_t]]
    path: str = None

    def __post_init__(self, sections: Sequence[section_t], /) -> None:
        """
        Raising exceptions is adapted here since execution cannot proceed without a
        valid specification.
        """
        specific = []
        for section in sections:
            as_dict = dict(
                (_elm.name, getattr(section, _elm.name))
                for _elm in dtcl.fields(section)
            )
            if section.controller is None:
                del as_dict["controller"]
                del as_dict["alternatives"]
                section = free_section_t(**as_dict)
            else:
                section = controlled_section_t(**as_dict)
            specific.append(section)

        self.extend(specific)
        self.path = SpecificationPath(sections)

        for section in specific:
            if isinstance(section, free_section_t):
                continue

            controller = section.controller
            if self.GetController(controller).type is ANY_TYPE:
                choices = choices_t.NewAnnotatedType(section.controlling_values)
                self.GetController(controller).type = type_t.NewFromTypeHint(choices)

        _SignalIssues(self)
        ISSUE_MANAGER.Report(with_time=True, how="rich")

    def AddUnitSection(self) -> None:
        """"""
        section = free_section_t(
            name=INI_UNIT_SECTION,
            definition="Unit definitions",
            description=f"Units can be used in any other section "
            f"to specify a parameter value as follows: "
            f"numerical_value{INI_UNIT_SEPARATOR}unit, e.g., 1.5'mm.",
            basic=True,
            optional=True,
            category=INI_UNIT_SECTION,
            is_growable=True,
        )
        self.append(section)

    def AddPluginParameter(
        self,
        section: str | any_section_h,
        name: str,
        /,
        *,
        definition: str = "Programmatic plugin parameter",
        description: str = "This parameter is not part of the specification. "
        'It was added programmatically in a "plugin" way.',
        basic: bool = True,
        stripe: any_hint_h | type_t = None,
        default: Any = MISSING_REQUIRED_VALUE,
        controlling_value: str = None,
    ) -> None:
        """
        See definition and description above.
        """
        if isinstance(section, str):
            section = self[section]

        parameter = parameter_t(
            name=name,
            definition=definition,
            description=description,
            basic=basic,
            type=stripe,
            default=default,
        )
        self._AddParameter(
            section, parameter, controlling_value, config_t.AddPluginParameter.__name__
        )

    def AddRuntimeParameter(
        self,
        section: str | any_section_h,
        name: str,
        value: str | float,
        /,
        *,
        controlling_value: str = None,
    ) -> parameter_t:
        """
        See definition and description below.
        The existence of such a method is justified by the fact that the parameter
        created can be a "normal" parameter or a unit depending on the section.

        value: float if unit.

        /!\ Cannot deal with str values with units.
        """
        if isinstance(section, str):
            section = self[section]

        if IsUnitSection(section.name):
            parameter_or_unit_t = unit_t
            basic = True
            converted = value
            stripe = None  # Correctly set by the constructor.
        else:
            parameter_or_unit_t = parameter_t
            basic = section.basic
            converted, _ = ObjectFromStr(value)
            stripe = type(converted)
        definition = "Programmatic runtime parameter"
        description = (
            "This parameter is not part of the specification. "
            "It was added programmatically because it was found in the INI document, "
            "or passed as a command-line argument."
        )
        parameter = parameter_or_unit_t(
            name=name,
            definition=definition,
            description=description,
            basic=basic,
            type=stripe,
            default=converted,  # Just a trick to prevent error if basic is False.
        )
        self._AddParameter(
            section, parameter, controlling_value, config_t.AddRuntimeParameter.__name__
        )

        return parameter

    def _AddParameter(
        self,
        section: any_section_h,
        parameter: parameter_t,
        controlling_value: str,
        caller: str,
        /,
    ) -> None:
        """"""
        if not section.is_growable:
            raise RuntimeError(
                f"{section.name}.{parameter.name}: "
                f"Attempt to add an unspecified parameter to a section accepting none."
            )

        if isinstance(section, free_section_t):
            section.append(parameter)
            return

        controller = section.controller
        if controlling_value is None:
            raise ValueError(
                f"{caller}: A Controlling value must be passed for parameter "
                f"{section.name}.{parameter.name}."
            )

        if controller.primary_value is None:
            # controller_t is immutable, so it must be re-created.
            section.controller = controller_t(
                section=controller.section,
                parameter=controller.parameter,
                primary_value=controlling_value,
            )
            controller = section.controller
            section[controlling_value] = parameters_t((parameter,))
            should_update_controller_choices = True
        elif controlling_value in section:
            section[controlling_value].append(parameter)
            should_update_controller_choices = False
        else:
            section[controlling_value] = parameters_t((parameter,))
            should_update_controller_choices = True

        if should_update_controller_choices:
            nnts = self.GetController(controller).type.annotations
            if nnts.__len__() > 0:
                first_annotation = nnts[0]
            else:
                first_annotation = None
            if (first_annotation is None) or isinstance(first_annotation, choices_t):
                choices = choices_t.NewAnnotatedType(section.controlling_values)
                new_type = type_t.NewFromTypeHint(choices)
            else:  # Must be callable_t.
                callable_ = callable_t.NewAnnotatedType(
                    kind=first_annotation.kind,
                    catalog=section.controlling_values,
                    allow_external=first_annotation.allow_external,
                )
                new_type = type_t.NewFromTypeHint(callable_)
            self.GetController(controller).type = new_type

    @property
    def section_names(self) -> Sequence[str]:
        """"""
        return tuple(_sct.name for _sct in self)

    def GetController(self, controller: controller_t, /) -> parameter_t:
        """"""
        return self[controller.section][controller.parameter]

    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(_elm.name == key for _elm in self)

    def __getitem__(self, key: str, /) -> any_section_h:
        """"""
        for section in self:
            if section.name == key:
                return section

        raise KeyError(f"{key}: Not a section of config.")

    def AsDict(self) -> dict[str, dict[str, Any]]:
        """"""
        return {_sct.name: _sct.AsDict() for _sct in self}

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            f"    [blue]path[/]={self.path}" f"[yellow]:{type(self.path).__name__}[/]",
        ]

        for section in self:
            output.append(text.indent(section.__rich__(), "    "))

        return "\n".join(output)


def _SignalIssues(self: config_t, /) -> None:
    """"""
    with ISSUE_MANAGER.AddedContextLevel(f'Specification "{self.path}"'):
        if self.__len__() == 0:
            ISSUE_MANAGER.Add(f"Empty specification")
            return

        names = self.section_names
        if names.__len__() > set(names).__len__():
            ISSUE_MANAGER.Add("Repeated section names")

        for section in self:
            if isinstance(section, controlled_section_t):
                controller = section.controller
                if controller.section not in self:
                    ISSUE_MANAGER.Add(
                        f'Unspecified parameter '
                        f'"{controller.section}.{controller.parameter}" declared as '
                        f'controller of section "{section.name}"'
                    )
                else:
                    controller_section = self[controller.section]
                    if isinstance(controller_section, controlled_section_t):
                        ISSUE_MANAGER.Add(
                            f'Section "{controller.section}" of parameter '
                            f'"{controller.parameter}", which controls section '
                            f'"{section.name}", is itself controlled'
                        )
                    if controller.parameter not in controller_section:
                        ISSUE_MANAGER.Add(
                            f'Unspecified parameter '
                            f'"{controller.section}.{controller.parameter}" declared '
                            f'as controller of section "{section.name}"'
                        )
                    else:
                        controller_section: free_section_t
                        controller_parameter = controller_section[controller.parameter]
                        if controller_parameter.optional:
                            ISSUE_MANAGER.Add(
                                f'Optional parameter '
                                f'"{controller.section}.{controller.parameter}" '
                                f'declared as controller of section "{section.name}"'
                            )



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import ClassVar

from issue_manager import ISSUE_MANAGER
from str_to_obj import annotation_t

number_h = int | float


@dtcl.dataclass(slots=True, repr=False, eq=False)
class number_t(annotation_t):
    INFINITY_NEG: ClassVar[float] = -float("inf")
    INFINITY_POS: ClassVar[float] = float("inf")
    INFINITE_PRECISION: ClassVar[float] = 0.0

    ACCEPTED_TYPES = (int, float)
    min: number_h = INFINITY_NEG
    max: number_h = INFINITY_POS
    min_inclusive: bool = True
    max_inclusive: bool = True
    precision: number_h = INFINITE_PRECISION

    def __post_init__(self) -> None:
        """"""
        stripe = self.__class__

        with ISSUE_MANAGER.AddedContextLevel("Number Annotation"):
            if (self.min != stripe.INFINITY_NEG) and not isinstance(
                self.min, number_h.__args__
            ):
                ISSUE_MANAGER.Add(
                    f"Invalid type for min value {self.min}",
                    actual=type(self.min).__name__,
                    expected=number_h,
                )
            if (self.max != stripe.INFINITY_POS) and not isinstance(
                self.max, number_h.__args__
            ):
                ISSUE_MANAGER.Add(
                    f"Invalid type for max value {self.max}",
                    actual=type(self.max).__name__,
                    expected=number_h,
                )
            if (self.precision != stripe.INFINITE_PRECISION) and not isinstance(
                self.precision, number_h.__args__
            ):
                ISSUE_MANAGER.Add(
                    f"Invalid type for precision {self.precision}",
                    actual=type(self.precision).__name__,
                    expected=number_h,
                )
            if self.precision < 0:
                ISSUE_MANAGER.Add(f"Invalid, negative precision {self.precision}")
            if (self.min > self.max) or (
                (self.min == self.max)
                and not (self.min_inclusive and self.max_inclusive)
            ):
                if self.min_inclusive:
                    opening = "["
                else:
                    opening = "]"
                if self.max_inclusive:
                    closing = "]"
                else:
                    closing = "["
                ISSUE_MANAGER.Add(
                    f"Empty value interval {opening}{self.min},{self.max}{closing}"
                )

    def ValueIsCompliant(self, value: number_h, /) -> list[str]:
        """"""
        issues = annotation_t.ValueIsCompliant(self, value)
        if issues.__len__() > 0:
            return issues

        if self.min <= value <= self.max:
            if ((value == self.min) and not self.min_inclusive) or (
                (value == self.max) and not self.max_inclusive
            ):
                return [f"{value}: Value outside prescribed interval."]

            if (self.precision != self.__class__.INFINITE_PRECISION) and (
                self.precision * int(value / self.precision) != value
            ):
                return [f"{value}: Value of higher precision than the prescribed one."]

            return []
        else:
            return [f"{value}: Value outside prescribed interval."]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import itertools as ittl
import textwrap as text
from typing import Any

from str_to_obj.interface.console import TypeAsRichStr

from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.phase.specification.base import base_t
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.section.controller import controller_t


@dtcl.dataclass(repr=False, eq=False)
class _section_t(base_t):
    """
    Default values only help complying with the constraint: no mandatory attributes
    after the default ones in base_t. In practice, they are mandatory.
    """

    category: str = ""
    optional: bool = False
    is_growable: bool = False

    def __rich_intro__(self) -> list[str]:
        """"""
        return [
            TypeAsRichStr(self),
            *text.indent(super().__rich__(), "    ").splitlines(),
            f"    [blue]Category[/]@=@{self.category}",
            f"    [blue]Optional[/]@=@{self.optional}",
            f"    [blue]Growable[/]@=@{self.is_growable}",
        ]

    @staticmethod
    def __rich_final__(elements: list[str], /) -> str:
        """"""
        elements = AlignedOnSeparator(elements, "@=@", " = ")
        return "\n".join(elements)


@dtcl.dataclass(init=False, repr=False, eq=False)
class parameters_t(list[parameter_t]):
    def __contains__(self, key: str, /) -> bool:
        """"""
        return any(_elm.name == key for _elm in self)

    def __getitem__(self, key: str, /) -> parameter_t:
        """"""
        for parameter in self:
            if parameter.name == key:
                return parameter

        raise KeyError(f"{key}: Not a valid parameter.")


@dtcl.dataclass(slots=True, repr=False, eq=False)
class free_section_t(_section_t, parameters_t):
    parameters: dtcl.InitVar[list[parameter_t] | None] = None

    def __post_init__(self, parameters: list[parameter_t] | None, /) -> None:
        """"""
        if parameters is not None:
            self.extend(parameters)

    @property
    def all_parameters(self) -> parameters_t:
        """"""
        return self

    def ActiveParameters(self, _: str, /) -> parameters_t:
        """"""
        return self

    def AsDict(self) -> dict[str, Any]:
        """"""
        return {_prm.name: _prm.default for _prm in self if _prm.optional}

    def __rich__(self) -> str:
        """"""
        output = self.__rich_intro__()
        output.extend(text.indent(_prm.__rich__(), "    ") for _prm in self)
        return self.__rich_final__(output)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class controlled_section_t(_section_t, dict[str, parameters_t]):
    """
    Default value only help complying with the constraint: no mandatory attributes
    after the default ones in base_t. In practice, it is mandatory.
    """

    controller: controller_t = None
    parameters: dtcl.InitVar[list[parameter_t] | None] = None
    alternatives: dtcl.InitVar[dict[str, list[parameter_t]] | None] = None

    def __post_init__(
        self,
        parameters: list[parameter_t] | None,
        alternatives: dict[str, list[parameter_t]] | None,
        /,
    ) -> None:
        """"""
        if parameters is None:
            return

        self[self.controller.primary_value] = parameters_t(parameters)
        for key, value in alternatives.items():
            self[key] = parameters_t(value)

    @property
    def controlling_values(self) -> tuple[str, ...]:
        """"""
        return tuple(self.keys())

    @property
    def all_parameters(self) -> parameters_t:
        """"""
        return parameters_t(ittl.chain(*self.values()))

    def ActiveParameters(self, controlling_value: str, /) -> parameters_t:
        """"""
        if self.__len__() > 0:
            return self[controlling_value]

        # Despite being a controlled section, there can be no alternatives if the
        # section was specified empty. In a GUI context, the section would have
        # been populated programmatically. In a CLI context, it is not.
        return parameters_t()

    def AsDict(self) -> dict[str, Any]:
        """"""
        output = {}

        for value, parameters in self.items():
            output.update(
                {
                    (value, _prm.name): _prm.default
                    for _prm in parameters
                    if _prm.optional
                }
            )

        return output

    def __rich__(self) -> str:
        """"""
        output = [
            *self.__rich_intro__(),
            f"    [blue]Controller[/]@=@{self.controller}",
            "    Alternatives:",
        ]

        for ctl_name, parameters in self.items():
            output.append(f"        {ctl_name}")
            for parameter in parameters:
                output.append(text.indent(parameter.__rich__(), "            "))

        return self.__rich_final__(output)


any_section_h = free_section_t | controlled_section_t
